import json
from typing import Any, Dict, Iterator, List, Optional

from loguru import logger

from .local_metadata_source import ColumnMetadata, LocalMetadataSource

try:
    from avro.datafile import DataFileReader, DataFileWriter
    from avro.io import DatumReader, DatumWriter

    AVRO_INSTALLED = True

except ImportError:
    logger.warning("AVRO optional dependency is not installed.")
    AVRO_INSTALLED = False

if AVRO_INSTALLED:

    class AvroSource(LocalMetadataSource):
        """Instance for a local Avro file."""

        def read(self) -> DataFileReader:
            """Read the AVRO file."""
            with open(self.local_path, "rb") as file:
                return DataFileReader(file, DatumReader())

        def get_field_attribute(
            self, attribute_name: str
        ) -> List[Optional[ColumnMetadata]]:
            """
            Get the specific attribute from the AVRO Schema.

            :param attribute_name: the attribute name to get
            :return: the list of attributes in the fields
            """
            reader = self.read()
            schema = json.loads(reader.meta["avro.schema"])
            return [
                ColumnMetadata(column_name=str(field[attribute_name]))
                if attribute_name in field
                else None
                for field in schema["fields"]
            ]

        def get_column_names(self) -> Iterator[ColumnMetadata]:
            """
            Get column names from the AVRO file.

            :return: the list of the column names
            """
            reader = self.read()
            schema = json.loads(reader.meta["avro.schema"])
            for field in schema["fields"]:
                yield ColumnMetadata(column_name=field["name"])

        @property
        def namespace(self) -> str:
            """
            Namespace of the AVRO schema.

            :return: the namespace
            """
            reader = self.read()
            schema = json.loads(reader.meta["avro.schema"])
            return schema["namespace"]

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name o of the source.
            """
            return "Avro"



from abc import ABC, abstractmethod
from typing import Iterator, List, Optional

from pydantic import BaseModel


class Metadata(BaseModel, ABC):
    """Metadata contract."""

    @abstractmethod
    def as_list(self) -> Iterator[str]:
        """
        Return as a raw list of strings.

        :return: a list of string
        """
        pass


class ColumnMetadata(Metadata):
    """Column Metadata instance."""

    column_name: str
    column_comment: Optional[str] = None

    def as_list(self) -> Iterator[str]:
        """
        Return as a raw list of strings.

        :return: a list of string
        """
        yield self.column_name
        if self.column_comment:
            yield self.column_comment


class MetadataSource(BaseModel, ABC):
    """Metadata Source contract."""

    @classmethod
    def type(cls) -> str:
        """
        The type of the source.

        :return: the name of the source.
        """
        pass



from typing import List, NamedTuple, Optional, Tuple

from pydantic import BaseModel, Field
from rich.console import Console
from rich.markup import escape
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
)
from rich.table import Table

from .data_rules import MetadataGuardianResults


class ProgressionBar(Progress):
    """
    Progression Bar provides a progression bar to display the results of the scanner.
    """

    task_id: Optional[TaskID] = None

    def __init__(self, disable: bool) -> None:
        super().__init__(
            SpinnerColumn(),
            "[progress.description]{task.description}: [red]{task.fields[current_item]}",
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}% ({task.completed}/{task.total})-",
            TimeRemainingColumn(),
            disable=disable,
        )

    def __enter__(self) -> "ProgressionBar":
        super().__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        if self.task_id is not None:
            super().update(self.task_id, current_item="Done")
        super().__exit__(exc_type, exc_val, exc_tb)

    def add_task_with_item(
        self,
        item_name: str,
        source_type: str,
        total: int,
        current_item: str = "Starting",
    ) -> None:
        """
        Add task in the Progression Bar.

        :param item_name: the name of the item to search
        :param current_item: the name of the current item
        :param source_type: the source type
        :param total: total of the number of tables
        :return: the created Task
        """
        task_details = f"[{item_name}]" if item_name else ""
        task_description = f"[bold cyan]Searching in the {escape(source_type)} metadata source{escape(task_details)}"
        task_id = super().add_task(
            description=task_description,
            total=total,
            current_item=escape(current_item),
        )
        self.task_id = task_id

    def update_item(self, current_item: str) -> None:
        """
        Update the current item of the task.

        :param current_item: the name of the current item
        :return:
        """
        if self.task_id is not None:
            super().update(self.task_id, advance=1, current_item=current_item)


class ReportResults(BaseModel):
    """Metadata Guardian Results."""

    source: str
    results: List[MetadataGuardianResults] = Field(default_factory=list)


class MetadataGuardianReport(BaseModel):
    """Metadata Guardian Report."""

    report_results: List[ReportResults] = Field(default_factory=list)

    def append(self, other_report: "MetadataGuardianReport") -> None:
        """
        Concat the results before making the report.

        :param other_report: other report to append
        :return:
        """
        self.report_results = self.report_results + other_report.report_results

    def to_console(self) -> None:
        """
        Display the metadata guardian results to the console.

        :return:
        """
        _console = Console()
        _table = Table(
            title=":magnifying_glass_tilted_right: Metadata Guardian report",
            show_header=True,
            header_style="bold",
            show_lines=True,
        )
        _table.add_column("Category", style="yellow", no_wrap=True)
        _table.add_column("Source", style="cyan", no_wrap=True)
        _table.add_column("Content", style="cyan", no_wrap=True)
        _table.add_column("Name", style="magenta", no_wrap=True)
        _table.add_column("Documentation")
        for report in self.report_results:
            for result in report.results:
                for data_rule in result.data_rules:
                    _table.add_row(
                        result.category,
                        report.source,
                        result.content.strip(),
                        data_rule.rule_name,
                        data_rule.documentation,
                    )
        if _table.rows:
            _console.print(
                f":exclamation: Metadata Guardian detected {len(_table.rows)} data rules violations."
            )
            _console.print(_table)
        else:
            _console.print(
                f":thumbs_up: No data rules violation were detected by Metadata Guardian."
            )



from typing import Any, Dict, Iterator, List, Optional, Text, Union

from loguru import logger

from .local_metadata_source import ColumnMetadata, LocalMetadataSource

try:
    from avro.schema import parse

    AVRO_INSTALLED = True

except ImportError:
    logger.warning("AVRO optional dependency is not installed.")
    AVRO_INSTALLED = False

if AVRO_INSTALLED:

    class AvroSchemaSource(LocalMetadataSource):
        """Instance for a local Avro Schema file."""

        def read(self) -> Union[Text, bytes]:
            """Read the AVRO Schema file."""
            with open(self.local_path, "r") as file:
                return file.read()

        def get_column_names(self) -> Iterator[ColumnMetadata]:
            """
            Get column names from the AVRO Schema file.

            :return: the list of the column names
            """
            schema = parse(self.read())
            for field in schema.fields:
                yield ColumnMetadata(column_name=field.name)

        @property
        def namespace(self) -> str:
            """
            Namespace of the AVRO schema.

            :return: the namespace
            """
            schema = parse(self.read())
            return schema.namespace

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name o of the source.
            """
            return "AvroSchema"



import asyncio
import json
from typing import Any, List, Optional

import typer
from loguru import logger

from ... import ColumnScanner, DataRules
from ...source.external.external_metadata_source import ExternalMetadataSource

app = typer.Typer()


def get_external_source(source: str, configuration) -> ExternalMetadataSource:  # type: ignore
    sources = list(displayed=False)
    if source not in sources:
        raise ValueError(f"This source is not available in the list: {sources}")

    try:
        selected_source = next(
            cls
            for cls in ExternalMetadataSource.__subclasses__()
            if cls.type() == source
        )
    except Exception as exception:
        logger.exception("This source initiation failed.")
        raise exception
    return selected_source.parse_raw(configuration)


@app.command(help="List the external metadata sources")
def list(displayed: bool = True) -> List[str]:
    sources = [cls.type() for cls in ExternalMetadataSource.__subclasses__()]
    if displayed:
        logger.info(f"Available External sources: {sources}")
    return sources


@app.command(help="Scan async the external metadata sources with the ColumnScanner")
# type: ignore
def scan_async(
    external_source: str,
    database_name: str,
    data_rules_path: str,
    configuration,
    table_name: Optional[str] = None,
    include_comments: bool = False,
) -> None:
    source = get_external_source(source=external_source, configuration=configuration)

    data_rules = DataRules.from_path(path=data_rules_path)
    column_scanner = ColumnScanner(
        data_rules=data_rules, progression_bar_disabled=False
    )
    with source:
        with source:
            report = asyncio.run(
                column_scanner.scan_external_async(
                    source,
                    database_name=database_name,
                    include_comment=include_comments,
                    table_name=table_name,
                )
            )
            report.to_console()


@app.command(help="Scan the external metadata sources with the ColumnScanner")
# type: ignore
def scan(
    external_source: str,
    database_name: str,
    data_rules_path: str,
    configuration,
    table_name: Optional[str] = None,
    include_comments: bool = False,
) -> None:
    source = get_external_source(source=external_source, configuration=configuration)

    data_rules = DataRules.from_path(path=data_rules_path)
    column_scanner = ColumnScanner(
        data_rules=data_rules, progression_bar_disabled=False
    )
    with source:
        report = column_scanner.scan_external(
            source,
            database_name=database_name,
            table_name=table_name,
            include_comment=include_comments,
        )
        report.to_console()



import argparse
import os

from metadata_guardian import (
    AvailableCategory,
    ColumnScanner,
    DataRules,
    ExternalMetadataSource,
)
from metadata_guardian.source import (
    AthenaSource,
    BigQuerySource,
    DeltaTableSource,
    GlueSource,
    KafkaSchemaRegistrySource,
    MySQLSource,
    SnowflakeSource,
)


def get_snowflake() -> ExternalMetadataSource:
    return SnowflakeSource(
        sf_account=os.environ["SNOWFLAKE_ACCOUNT"],
        sf_user=os.environ["SNOWFLAKE_USER"],
        sf_password=os.environ["SNOWFLAKE_PASSWORD"],
        warehouse=os.environ["SNOWFLAKE_WAREHOUSE"],
        schema_name=os.environ["SNOWFLAKE_SCHEMA_NAME"],
    )


def get_gcp_bigquery() -> ExternalMetadataSource:
    return BigQuerySource(
        service_account_json_path=os.environ["BIGQUERY_SERVICE_ACCOUNT"],
        project=os.environ["BIGQUERY_PROJECT"],
        location=os.environ["BIGQUERY_LOCATION"],
    )


def get_kafka_schema_registry() -> ExternalMetadataSource:
    return KafkaSchemaRegistrySource(url=os.environ["KAFKA_SCHEMA_REGISTRY_URL"])


def get_delta_table() -> ExternalMetadataSource:
    return DeltaTableSource(uri=os.environ["DELTA_TABLE_URI"])


def get_mysql() -> ExternalMetadataSource:
    return MySQLSource(
        user=os.environ["MYSQL_USER"],
        password=os.environ["MYSQL_PASSWORD"],
        host=os.environ["MYSQL_HOST"],
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-rules",
        choices=["PII", "INCLUSION"],
        default="PII",
        help="The Data Rules to use",
    )
    parser.add_argument(
        "--external-source",
        choices=[
            "Snowflake",
            "GCP BigQuery",
            "Kafka Schema Registry",
            "Delta Table",
            "MySQL",
        ],
        required=True,
        help="The External Metadata Source to use",
    )
    parser.add_argument(
        "--scanner", choices=["ColumnScanner"], help="The scanner to use"
    )
    parser.add_argument(
        "--database_name", required=True, help="The database name to scan"
    )
    parser.add_argument(
        "--include_comments", default=True, help="Include the comments in the scan"
    )
    args = parser.parse_args()
    data_rules = DataRules.from_available_category(
        category=AvailableCategory[args.data_rules]
    )
    column_scanner = ColumnScanner(
        data_rules=data_rules, progression_bar_disabled=False
    )

    if args.external_source == "Snowflake":
        source = get_snowflake()
    elif args.external_source == "GCP BigQuery":
        source = get_gcp_bigquery()
    elif args.external_source == "Kafka Schema Registry":
        source = get_kafka_schema_registry()
    elif args.external_source == "Delta Table":
        source = get_delta_table()
    elif args.external_source == "MySQL":
        source = get_mysql()

    with source:
        report = column_scanner.scan_external(
            source,
            database_name=args.database_name,
            include_comment=args.include_comments,
        )
        report.to_console()



import os

from rich.logging import RichHandler


def configure_logger() -> None:
    """
    Configure the loguru configuration with Rich.

    :return:
    """
    if "LOGURU_LEVEL" not in os.environ:
        os.environ["LOGURU_LEVEL"] = "INFO"
    from loguru import logger

    logger.configure(
        handlers=[{"sink": RichHandler(markup=True), "format": "{message}"}]
    )


configure_logger()



from unittest.mock import patch

from metadata_guardian.source import (
    ColumnMetadata,
    SnowflakeAuthenticator,
    SnowflakeSource,
)


@patch("snowflake.connector")
def test_snowflake_source_get_column_names(mock_connection):
    database_name = "test_database"
    schema_name = "PUBLIC"
    table_name = "test_table"
    sf_account = "sf_account"
    sf_user = "sf_user"
    sf_password = "sf_password"
    warehouse = "warehouse"
    mocked_cursor_one = mock_connection.connect().cursor.return_value
    mocked_cursor_one.description = [["name"], ["phone"]]
    mocked_cursor_one.fetchall.return_value = [
        (database_name, table_name, "column1", "", "", "", "", "", "comment1"),
        (database_name, table_name, "column2", "", "", "", "", "", "comment2"),
    ]
    mocked_cursor_one.execute.call_args == f'SHOW COLUMNS IN "{database_name}"."{schema_name}"."{table_name}"'
    expected = [
        ColumnMetadata(column_name="column1", column_comment="comment1"),
        ColumnMetadata(column_name="column2", column_comment="comment2"),
    ]

    source = SnowflakeSource(
        sf_account=sf_account,
        sf_user=sf_user,
        sf_password=sf_password,
        warehouse=warehouse,
        schema_name=schema_name,
    )

    column_names = source.get_column_names(
        database_name=database_name, table_name=table_name, include_comment=True
    )

    assert list(column_names) == expected
    assert source.authenticator == SnowflakeAuthenticator.USER_PWD


@patch("snowflake.connector")
def test_snowflake_source_get_table_names_list(mock_connection):
    database_name = "test_database"
    schema_name = "PUBLIC"
    sf_account = "sf_account"
    sf_user = "sf_user"
    sf_password = "sf_password"
    warehouse = "warehouse"
    mocked_cursor_one = mock_connection.connect().cursor.return_value
    mocked_cursor_one.fetchall.return_value = [
        (database_name, "TEST_TABLE"),
        (database_name, "TEST_TABLE2"),
    ]
    mocked_cursor_one.execute.call_args == f'SHOW TABLES IN DATABASE "{database_name}"'
    expected = ["TEST_TABLE", "TEST_TABLE2"]

    source = SnowflakeSource(
        sf_account=sf_account,
        sf_user=sf_user,
        sf_password=sf_password,
        warehouse=warehouse,
        schema_name=schema_name,
    )

    table_names = source.get_table_names_list(database_name=database_name)

    assert list(table_names) == expected
    assert source.authenticator == SnowflakeAuthenticator.USER_PWD



from typing import List

import typer
from loguru import logger

from ... import ColumnScanner, DataRules
from ...source.local.local_metadata_source import LocalMetadataSource

app = typer.Typer()


def get_local_source(source: str, path: str) -> LocalMetadataSource:
    sources = list(displayed=False)
    if source not in sources:
        raise ValueError(f"This source is not available in the list: {sources}")

    try:
        selected_source = next(
            cls for cls in LocalMetadataSource.__subclasses__() if cls.type() == source
        )(local_path=path)
    except Exception as exception:
        logger.exception("This source initiation failed.")
        raise exception
    return selected_source


@app.command(help="List the local metadata sources")
def list(displayed: bool = True) -> List[str]:
    sources = [cls.type() for cls in LocalMetadataSource.__subclasses__()]
    if displayed:
        logger.info(f"Available External sources: {sources}")
    return sources


@app.command(help="Scan the local metadata sources with the ColumnScanner")
def scan(local_source: str, data_rules_path: str, path: str) -> None:
    source = get_local_source(source=local_source, path=path)

    data_rules = DataRules.from_path(path=data_rules_path)
    column_scanner = ColumnScanner(
        data_rules=data_rules, progression_bar_disabled=False
    )
    report = column_scanner.scan_local(source=source)
    report.to_console()



from typing import Any, Dict, Iterator, List, Optional

from pyarrow.orc import ORCFile

from .local_metadata_source import ColumnMetadata, LocalMetadataSource


class ORCSource(LocalMetadataSource):
    """Instance for a local ORC file."""

    def read(self) -> ORCFile:
        """
        Read the ORC file.

        :return:
        """
        return ORCFile(self.local_path)

    def get_column_names(self) -> Iterator[ColumnMetadata]:
        """
        Get the column names from the schema.

        :return: the list of the column names
        """
        for column_name in self.read().schema.names:
            yield ColumnMetadata(column_name=column_name)

    @classmethod
    def type(cls) -> str:
        """
        The type of the source.

        :return: the name of the source.
        """
        return "ORC"



import pytest

from metadata_guardian.source import ColumnMetadata
from metadata_guardian.source.local.orc_source import ORCSource


@pytest.mark.parametrize("local_file", ["example.orc"], indirect=["local_file"])
def test_orc_source(local_file):
    source = ORCSource(local_path=local_file)
    expected = [
        ColumnMetadata(column_name="boolean1"),
        ColumnMetadata(column_name="byte1"),
        ColumnMetadata(column_name="short1"),
        ColumnMetadata(column_name="int1"),
        ColumnMetadata(column_name="long1"),
        ColumnMetadata(column_name="float1"),
        ColumnMetadata(column_name="double1"),
        ColumnMetadata(column_name="bytes1"),
        ColumnMetadata(column_name="string1"),
        ColumnMetadata(column_name="middle"),
        ColumnMetadata(column_name="list"),
        ColumnMetadata(column_name="map"),
    ]

    column_names = source.get_column_names()

    assert list(column_names) == expected



from unittest.mock import Mock, patch

from deltalake import Field, Schema

from metadata_guardian.source import ColumnMetadata, DeltaTableSource


@patch("deltalake.DeltaTable")
def test_deltatable_source_get_column_names(mock_connection):
    uri = "s3://test_table"
    schema = Schema(
        fields=[
            Field(
                name="timestamp",
                type="string",
                nullable=False,
                metadata={"comment": "comment1"},
            ),
            Field(
                name="address_id",
                type="int",
                nullable=False,
                metadata={"comment": "comment2"},
            ),
        ],
        json_value={},
    )
    mock_connection.schema.return_value = schema
    expected = [
        ColumnMetadata(
            column_name="timestamp", column_comment="{'comment': 'comment1'}"
        ),
        ColumnMetadata(
            column_name="address_id", column_comment="{'comment': 'comment2'}"
        ),
    ]

    delta_table = DeltaTableSource(uri=uri)
    delta_table._connection = mock_connection
    column_names = delta_table.get_column_names(include_comment=True)

    assert list(column_names) == expected


@patch("deltalake.DeltaTable.from_data_catalog")
def test_deltatable_source_get_column_names_from_database_and_table(mock_connection):
    uri = "s3://test_table"
    database_name = "database_name"
    table_name = "table_name"
    external_data_catalog_disable = False
    schema = Schema(
        fields=[
            Field(
                name="timestamp",
                type="string",
                nullable=False,
                metadata={"comment": "comment1"},
            ),
            Field(
                name="address_id",
                type="int",
                nullable=False,
                metadata={"comment": "comment2"},
            ),
        ],
        json_value={},
    )
    mock_connection.return_value = mock_connection
    mock_connection.schema.return_value = schema
    expected = [
        ColumnMetadata(
            column_name="timestamp", column_comment="{'comment': 'comment1'}"
        ),
        ColumnMetadata(
            column_name="address_id", column_comment="{'comment': 'comment2'}"
        ),
    ]

    column_names = DeltaTableSource(
        uri=uri, external_data_catalog_disable=external_data_catalog_disable
    ).get_column_names(
        database_name=database_name, table_name=table_name, include_comment=True
    )

    assert list(column_names) == expected



from metadata_guardian.cli.external import get_external_source
from metadata_guardian.cli.local import get_local_source
from metadata_guardian.source import MySQLSource, ParquetSource


def test_get_external_source_():
    source = "MySQL"
    configuration = (
        '{"user": "user", "password": "password", "host": "host",'
        ' "extra_connection_args": {"test": "True"}}'
    )
    expected = MySQLSource.parse_raw(configuration)

    source = get_external_source(source=source, configuration=configuration)

    assert source == expected


def test_get_local_source_():
    source = "Parquet"
    path = "path"
    expected = ParquetSource(**{"local_path": path})

    source = get_local_source(source=source, path=path)

    assert source == expected



from unittest.mock import patch

from metadata_guardian.source import ColumnMetadata, MySQLAuthenticator, MySQLSource


@patch("pymysql.connect")
def test_mysql_source_get_column_names(mock_connection):
    database_name = "test"
    table_name = "test_table"
    user = "user"
    host = "localhost"
    password = "password"
    mock_connection.cursor.return_value = mock_connection
    mock_connection.fetchall.return_value = [
        {
            "Field": "words",
            "Type": "varchar(45)",
            "Collation": "utf8_general_ci",
            "Null": "YES",
            "Key": "",
            "Default": None,
            "Extra": "",
            "Privileges": "select,insert,update,references",
            "Comment": "Use column to contain words",
        },
        {
            "Field": "name",
            "Type": "varchar(45)",
            "Collation": "utf8_general_ci",
            "Null": "YES",
            "Key": "",
            "Default": None,
            "Extra": "",
            "Privileges": "select,insert,update,references",
            "Comment": "",
        },
    ]
    mock_connection.execute.call_args == f"SHOW FULL COLUMNS FROM {database_name}.{table_name}"
    expected = [
        ColumnMetadata(
            column_name="words", column_comment="Use column to contain words"
        ),
        ColumnMetadata(column_name="name"),
    ]

    source = MySQLSource(
        host=host,
        user=user,
        password=password,
    )
    source._connection = mock_connection

    column_names = source.get_column_names(
        database_name=database_name, table_name=table_name, include_comment=True
    )

    assert list(column_names) == expected
    assert source.authenticator == MySQLAuthenticator.USER_PWD


@patch("pymysql.connect")
def test_mysql_source_get_table_names_list(mock_connection):
    database_name = "test"
    user = "user"
    host = "localhost"
    password = "password"
    mock_connection.cursor.return_value = mock_connection
    mock_connection.fetchall.return_value = [
        {"Tables_in_test": "t1"},
        {"Tables_in_test": "t2"},
    ]
    mock_connection.execute.call_args == f"SHOW TABLES IN {database_name}"
    expected = ["t1", "t2"]

    source = MySQLSource(
        host=host,
        user=user,
        password=password,
    )
    source._connection = mock_connection

    table_names = source.get_table_names_list(database_name=database_name)

    assert list(table_names) == expected
    assert source.authenticator == MySQLAuthenticator.USER_PWD



import importlib.resources
from enum import Enum
from typing import Any, List

from loguru import logger
from pydantic import BaseModel, PrivateAttr

from .metadata_guardian import RawDataRule, RawDataRules


class AvailableCategory(Enum):
    """Available Data Rules Categories."""

    PII = "pii_rules.yaml"
    INCLUSION = "inclusion_rules.yaml"


class DataRule(BaseModel):
    """DataRule instance with a regex pattern and a documentation."""

    rule_name: str
    regex_pattern: str
    documentation: str


class MetadataGuardianResults(BaseModel):
    """Metadata Guardian Results instance with the content that matches with the data rules."""

    category: str
    content: str
    data_rules: List[DataRule]


class DataRules(BaseModel):
    """Data Rules instances."""

    _data_rules: Any = PrivateAttr()

    def __init__(self, data_rules: RawDataRules, **data: Any) -> None:
        super().__init__(**data)
        self._data_rules = data_rules

    @classmethod
    def from_new_category(
        cls, category: str, data_rules: List[DataRule]
    ) -> "DataRules":
        """
        Create data rules from a given category and data rules.

        :param category: the category of the Data Rules
        :param data_rules: the list of data rule
        :return: the Data Rules instance
        """
        data_rules = RawDataRules(
            category=category,
            data_rules=[
                RawDataRule(
                    rule_name=data_rule.rule_name,
                    pattern=str(data_rule.regex_pattern),
                    documentation=data_rule.documentation,
                )
                for data_rule in data_rules
            ],
        )
        return cls(data_rules=data_rules)

    @classmethod
    def from_path(cls, path: str) -> "DataRules":
        """
        Get Data Rules from a path.

        :param path: the path of the yaml file
        :return: the Data Rules instance
        """
        data_rules = RawDataRules.from_path(path)
        return cls(data_rules=data_rules)

    @classmethod
    def from_available_category(cls, category: AvailableCategory) -> "DataRules":
        """
        Get Data Rules from an available category.

        :param category: the available category of the data rules
        :return: the Data Rules instance
        """
        logger.debug(f"Creating Data Rules from the category {category.value}")
        if not isinstance(category, AvailableCategory):
            raise ValueError("The category must be an instance of AvailableCategory")
        with importlib.resources.path(
            "metadata_guardian.rules", category.value
        ) as resource:
            path = str(resource)
            return cls.from_path(path=path)

    def validate_word(self, word: str) -> MetadataGuardianResults:
        """
        Validate a word with the data rules defined.

        :param word: the word to validate
        :return: the metadata guardian results
        """
        logger.debug(f"Validate the Data Rules with the word {word}")
        result = self._data_rules.validate_word(word=word)
        return MetadataGuardianResults(
            category=result._category,
            content=result._content,
            data_rules=[
                DataRule(
                    rule_name=data_rule.rule_name,
                    regex_pattern=data_rule.pattern,
                    documentation=data_rule.documentation,
                )
                for data_rule in result._data_rules
            ],
        )

    def validate_words(self, words: List[str]) -> List[MetadataGuardianResults]:
        """
        Validate a list of words with the data rules defined.

        :param words: the words to validate
        :return: the metadata guardian results
        """
        logger.debug(f"Validate the Data Rules with the words {words}")
        results = self._data_rules.validate_words(words=words)
        return [
            MetadataGuardianResults(
                category=result._category,
                content=result._content,
                data_rules=[
                    DataRule(
                        rule_name=data_rule.rule_name,
                        regex_pattern=data_rule.pattern,
                        documentation=data_rule.documentation,
                    )
                    for data_rule in result._data_rules
                ],
            )
            for result in results
        ]

    def validate_file(self, path: str) -> List[MetadataGuardianResults]:
        """
        Validate a file content with the data rules defined.

        :param path: the file path
        :return: the metadata guardian results
        """
        logger.debug(f"Validate the Data Rules in the path {path}")
        results = self._data_rules.validate_file(path)
        return [
            MetadataGuardianResults(
                category=result._category,
                content=result._content,
                data_rules=[
                    DataRule(
                        rule_name=data_rule.rule_name,
                        regex_pattern=data_rule.pattern,
                        documentation=data_rule.documentation,
                    )
                    for data_rule in result._data_rules
                ],
            )
            for result in results
        ]



from typing import Any, Dict, Iterator, List, Optional

from loguru import logger
from pydantic import Field

from ..metadata_source import ColumnMetadata
from .external_metadata_source import (
    ExternalMetadataSource,
    ExternalMetadataSourceException,
)

try:
    import boto3
    import botocore
    from mypy_boto3_athena.client import AthenaClient
    from mypy_boto3_glue.client import GlueClient

    AWS_INSTALLED = True
except ImportError:
    logger.debug("AWS optional dependency is not installed.")
    AWS_INSTALLED = False


if AWS_INSTALLED:

    class AthenaSource(ExternalMetadataSource):
        """Athena Source instance."""

        s3_staging_dir: str
        catalog_name: str = "AWSDataCatalog"
        region_name: Optional[str] = None
        aws_access_key_id: Optional[str] = None
        aws_secret_access_key: Optional[str] = None
        extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

        def create_connection(self) -> None:
            """
            Create Athena connection.
            :return:
            """
            self._connection = boto3.client(
                "athena",
                region_name=self.region_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                **self.extra_connection_args,
            )

        def close_connection(self) -> None:
            pass

        def get_column_names(
            self, database_name: str, table_name: str, include_comment: bool = False
        ) -> Iterator[ColumnMetadata]:
            """
            Get the column names from the table.

            :param database_name: the database name
            :param table_name: the table name
            :param include_comment: include the comment
            :return: the list of the column names
            """
            try:
                if not self._connection:
                    self.create_connection()
                response = self._connection.get_table_metadata(
                    CatalogName=self.catalog_name,
                    DatabaseName=database_name,
                    TableName=table_name,
                )
                for row in response["TableMetadata"]["Columns"]:
                    column_name = row["Name"]
                    column_comment = None
                    if include_comment:
                        if "Comment" in row:
                            column_comment = row["Comment"]
                    yield ColumnMetadata(
                        column_name=column_name, column_comment=column_comment
                    )
            except botocore.exceptions.ClientError as error:
                logger.exception(
                    f"Error in getting columns name from AWS Athena {database_name}.{table_name} for catalog {self.catalog_name}"
                )
                raise ExternalMetadataSourceException(error)

        def get_table_names_list(self, database_name: str) -> Iterator[str]:
            """
            Get the table names list from the database in AWS Athena.

            :param database_name: the database name
            :return: the list of the table names of the database
            """
            try:
                if not self._connection:
                    self.create_connection()
                response = self._connection.list_table_metadata(
                    CatalogName=self.catalog_name,
                    DatabaseName=database_name,
                )
                for table in response["TableMetadataList"]:
                    yield table["Name"]
                while "NextToken" in response:
                    response = self._connection.list_table_metadata(
                        CatalogName=self.catalog_name,
                        DatabaseName=database_name,
                        NextToken=response["NextToken"],
                    )
                    for table in response["TableMetadataList"]:
                        yield table["Name"]
            except botocore.exceptions.ClientError as exception:
                logger.exception(
                    f"Error in getting table names list from AWS Athena from the database {database_name} for catalog {self.catalog_name}"
                )
                raise ExternalMetadataSourceException(exception)

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name o of the source.
            """
            return "AWS Athena"

    class GlueSource(ExternalMetadataSource):
        """Glue Source instance."""

        region_name: Optional[str] = None
        aws_access_key_id: Optional[str] = None
        aws_secret_access_key: Optional[str] = None
        extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

        def create_connection(self) -> None:
            """
            Create the Glue connection.

            :return:
            """
            self._connection = boto3.client(
                "glue",
                region_name=self.region_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                **self.extra_connection_args,
            )

        def close_connection(self) -> None:
            pass

        def get_column_names(
            self, database_name: str, table_name: str, include_comment: bool = False
        ) -> Iterator[ColumnMetadata]:
            """
            Get the column names from AWS Glue table.

            :param database_name: the name of the database
            :param table_name: the name of the table
            :param include_comment: include the comments
            :return: the list of the column names
            """
            try:
                if not self._connection:
                    self.create_connection()
                response = self._connection.get_table(
                    DatabaseName=database_name, Name=table_name
                )
                for row in response["Table"]["StorageDescriptor"]["Columns"]:
                    column_name = row["Name"]
                    column_comment = None
                    if include_comment:
                        if "Comment" in row:
                            column_comment = row["Comment"]
                    yield ColumnMetadata(
                        column_name=column_name, column_comment=column_comment
                    )
            except botocore.exceptions.ClientError as exception:
                logger.exception(
                    f"Error in getting columns name from AWS Glue from the table {database_name}.{table_name}"
                )
                raise ExternalMetadataSourceException(exception)

        def get_table_names_list(self, database_name: str) -> Iterator[str]:
            """
            Get the table names list from the database in AWS Glue.

            :param database_name: the database name
            :return: the list of the table names of the database
            """
            try:
                if not self._connection:
                    self.create_connection()
                response = self._connection.get_tables(
                    DatabaseName=database_name,
                )
                for table in response["TableList"]:
                    yield table["Name"]
                while "NextToken" in response:
                    response = self._connection.get_tables(
                        DatabaseName=database_name, NextToken=response["NextToken"]
                    )
                    for table in response["TableList"]:
                        yield table["Name"]
            except botocore.exceptions.ClientError as error:
                logger.exception(
                    f"Error in getting table names list from AWS Glue from the database {database_name}"
                )
                raise error

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name of the source.
            """
            return "AWS Glue"



import asyncio
import os
from abc import ABC, abstractmethod
from typing import Optional

from loguru import logger
from pyarrow import cpu_count
from pydantic import BaseModel

from .data_rules import DataRules
from .report import MetadataGuardianReport, ProgressionBar, ReportResults
from .source import ExternalMetadataSource, LocalMetadataSource, MetadataSource


class Scanner(BaseModel, ABC):
    """
    Scanner Interface.
    """

    @abstractmethod
    def scan_local(self, source: LocalMetadataSource) -> MetadataGuardianReport:
        """
        Scan the local source.

        :param source: the LocalMetadataSource to scan
        :return: a Metadata Guardian report
        """
        pass

    @abstractmethod
    def scan_external(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source.

        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        pass

    @abstractmethod
    async def scan_external_async(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        tasks_limit: int,
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source asynchronously.

        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param tasks_limit: the limit of the tasks to run in parallel
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        pass


class ColumnScanner(Scanner):
    """Column Scanner instance."""

    data_rules: DataRules
    progression_bar_disabled: bool = True

    def scan_local(self, source: LocalMetadataSource) -> MetadataGuardianReport:
        """
        Scan the column names from the local source.

        :param source: the MetadataSource to scan
        :return: a Metadata Guardian report
        """
        logger.debug(
            f"[blue]Launch the metadata scanning of the local provider {source.type}"
        )
        with ProgressionBar(disable=self.progression_bar_disabled) as progression_bar:
            progression_bar.add_task_with_item(
                item_name=source.local_path,
                source_type=str(source.type),
                total=1,
            )
            words = [
                word
                for column_metadata in source.get_column_names()
                for word in column_metadata.as_list()
            ]
            report = MetadataGuardianReport(
                report_results=[
                    ReportResults(
                        source=source.local_path,
                        results=self.data_rules.validate_words(words=words),
                    )
                ]
            )
            progression_bar.update_item(current_item=source.local_path)
        return report

    def scan_external(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source using a table name or a database name.

        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        logger.debug(
            f"[blue]Launch the metadata scanning of the external provider {source.type} for {database_name}"
        )
        with ProgressionBar(disable=self.progression_bar_disabled) as progression_bar:
            if table_name:
                progression_bar.add_task_with_item(
                    item_name=database_name,
                    source_type=str(source.type),
                    total=1,
                    current_item=table_name,
                )
                words = [
                    word
                    for column_metadata in source.get_column_names(
                        database_name=database_name,
                        table_name=table_name,
                        include_comment=include_comment,
                    )
                    for word in column_metadata.as_list()
                ]
                report = MetadataGuardianReport(
                    report_results=[
                        ReportResults(
                            source=f"{database_name}.{table_name}",
                            results=self.data_rules.validate_words(words=words),
                        )
                    ]
                )
                progression_bar.update_item(current_item=table_name)
            else:
                report = MetadataGuardianReport()
                table_names_list = list(
                    source.get_table_names_list(database_name=database_name)
                )
                progression_bar.add_task_with_item(
                    item_name=database_name,
                    source_type=str(source.type),
                    total=len(table_names_list),
                )

                for table_name in table_names_list:
                    words = [
                        word
                        for column_metadata in source.get_column_names(
                            database_name=database_name,
                            table_name=table_name,
                            include_comment=include_comment,
                        )
                        for word in column_metadata.as_list()
                    ]
                    report.append(
                        MetadataGuardianReport(
                            report_results=[
                                ReportResults(
                                    source=f"{database_name}.{table_name}",
                                    results=self.data_rules.validate_words(words=words),
                                )
                            ]
                        )
                    )
                    progression_bar.update_item(current_item=table_name)
        return report

    async def scan_external_async(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        tasks_limit: int = cpu_count(),
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source using a table name or a database name.
        Note that it can generate multiple concurrent calls to your metadata source.

        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param tasks_limit: the limit of the tasks to run in parallel
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        semaphore = asyncio.Semaphore(tasks_limit)
        logger.debug(
            f"[blue]Launch asynchronously the metadata scanning of the external provider {source.type} for the database {database_name}"
        )

        async def async_validate_words(
            progression_bar: ProgressionBar, table_name: str
        ) -> ReportResults:
            async with semaphore:
                loop = asyncio.get_event_loop()
                columns_metadata = await loop.run_in_executor(
                    None,
                    source.get_column_names,
                    database_name,
                    table_name,
                    include_comment,
                )
                words = [
                    word
                    for column_metadata in columns_metadata
                    for word in column_metadata.as_list()
                ]
                progression_bar.update_item(current_item=table_name)
                return ReportResults(
                    source=f"{database_name}.{table_name}",
                    results=self.data_rules.validate_words(words=words),
                )

        with ProgressionBar(disable=self.progression_bar_disabled) as progression_bar:
            if table_name:
                tasks = [
                    async_validate_words(
                        progression_bar=progression_bar, table_name=table_name
                    )
                ]
            else:
                table_names_list = source.get_table_names_list(
                    database_name=database_name
                )

                tasks = [
                    async_validate_words(
                        progression_bar=progression_bar, table_name=table_name
                    )
                    for table_name in table_names_list
                ]
            progression_bar.add_task_with_item(
                item_name=database_name,
                source_type=str(source.type),
                total=len(tasks),
            )
            report_results = await asyncio.gather(*tasks)
            report = MetadataGuardianReport(report_results=report_results)
        return report


class ContentFilesScanner(BaseModel):
    """Content Files Scanner instance."""

    data_rules: DataRules
    progression_bar_disabled: bool = True

    def scan_local_file(self, path: str) -> MetadataGuardianReport:
        """
        Scan a file with data rules.

        :param path: the path of the file to scan
        :return: a Metadata Guardian report
        """
        logger.debug(
            f"[blue]Launch the metadata scanning the content of the file {path}"
        )
        progression_bar: ProgressionBar
        with ProgressionBar(disable=self.progression_bar_disabled) as progression_bar:
            progression_bar.add_task_with_item(
                item_name=path, source_type="files", total=1
            )
            report = MetadataGuardianReport(
                report_results=[
                    ReportResults(
                        source=path, results=self.data_rules.validate_file(path=path)
                    )
                ]
            )
            progression_bar.update_item(current_item=path)

        return report

    def scan_directory(
        self, directory_path: str, file_names_extension: str
    ) -> MetadataGuardianReport:
        """
        Scan all the files inside directory path with the file name extension.

        :param directory_path: the directory path to scan
        :param file_names_extension: the file name extension to include (without the ".")
        :return: a Metadata Guardian report
        """
        logger.debug(
            f"[blue]Launch the metadata scanning the content of the files {directory_path} with extension{file_names_extension}"
        )
        report = MetadataGuardianReport()
        for root, dirs, files in os.walk(directory_path):
            for name in files:
                path = f"{root}/{name}"
                if name.endswith(f".{file_names_extension}"):
                    report.append(other_report=self.scan_local_file(path=path))
        return report



from .local_metadata_source import LocalMetadataSource


class ParquetSource(LocalMetadataSource):
    """Instance for a local Parquet file."""

    @classmethod
    def type(cls) -> str:
        """
        The type of the source.

        :return: the name of the source.
        """
        return "Parquet"



from metadata_guardian.source import ColumnMetadata


def test_column_metadata():
    column_name = "column_name"
    column_comment = "column_comment"
    expected = [column_name, column_comment]

    column_metadata = ColumnMetadata(
        column_name=column_name, column_comment=column_comment
    )

    assert list(column_metadata.as_list()) == expected



from abc import abstractmethod
from typing import Any, Dict, Iterator, List, Optional

from loguru import logger
from pydantic import PrivateAttr

from ...exceptions import MetadataGuardianException
from ..metadata_source import ColumnMetadata, MetadataSource


class ExternalMetadataSource(MetadataSource):
    """ExternalMetadataSource Source."""

    _connection: Any = PrivateAttr()

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._connection = None

    def __enter__(self) -> "ExternalMetadataSource":
        try:
            self.create_connection()
        except Exception as exception:
            logger.exception(
                "Error raised while opening the Metadata Source connection"
            )
            raise exception
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> "ExternalMetadataSource":  # type: ignore
        try:
            self.close_connection()
        except Exception as exception:
            logger.exception(
                "Error raised while closing the Metadata Source connection"
            )
            raise exception
        return self

    @abstractmethod
    def get_column_names(
        self,
        database_name: str,
        table_name: str,
        include_comment: bool = False,
    ) -> Iterator[ColumnMetadata]:
        """
        Get the column names from the schema.

        :param database_name: the database name
        :param table_name: the table name
        :param include_comment: include the comment
        :return: the list of the column names
        """
        pass

    @abstractmethod
    def get_table_names_list(self, database_name: str) -> Iterator[str]:
        """
        Get the table names list from the database.

        :param database_name: the database name
        :return: the list of the table names of the database
        """
        pass

    @abstractmethod
    def create_connection(self) -> None:
        """
        Create the connection of the source.

        :return:
        """
        pass

    def close_connection(self) -> None:
        """
        Close the connection of the source.

        :return:
        """
        pass


class ExternalMetadataSourceException(MetadataGuardianException):
    """Raised where there is an exception to describe a external metadata source exception."""

    pass



from typing import Any, Dict, Iterator, List, Optional

from loguru import logger
from pydantic import Field

from ..metadata_source import ColumnMetadata
from .external_metadata_source import (
    ExternalMetadataSource,
    ExternalMetadataSourceException,
)

try:
    from deltalake import DataCatalog, DeltaTable

    DELTA_LAKE_INSTALLED = True
except ImportError:
    logger.debug("Delta Lake optional dependency is not installed.")
    DELTA_LAKE_INSTALLED = False

if DELTA_LAKE_INSTALLED:

    class DeltaTableSource(ExternalMetadataSource):
        uri: str
        data_catalog: DataCatalog = DataCatalog.AWS
        external_data_catalog_disable: bool = True
        extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

        def create_connection(self) -> None:
            """
            Create the DeltaTable instance.

            :return:
            """
            self._connection = DeltaTable(self.uri, **self.extra_connection_args)

        def close_connection(self) -> None:
            pass

        def get_column_names(
            self,
            database_name: Optional[str] = None,
            table_name: Optional[str] = None,
            include_comment: bool = False,
        ) -> Iterator[ColumnMetadata]:
            """
            Get column names from the Delta table.

            :param database_name: the database name
            :param table_name: the table name
            :param include_comment: include the comment
            :return: the list of the column names
            """
            try:
                if (
                    not self.external_data_catalog_disable
                    and database_name
                    and table_name
                ):
                    self._connection = DeltaTable.from_data_catalog(
                        data_catalog=self.data_catalog,
                        database_name=database_name,
                        table_name=table_name,
                    )
                elif not self._connection:
                    self.create_connection()
                schema = self._connection.schema()
                for field in schema.fields:
                    column_comment = None
                    if include_comment and field.metadata:
                        column_comment = str(field.metadata)
                    yield ColumnMetadata(
                        column_name=field.name, column_comment=column_comment
                    )
            except Exception as exception:
                logger.exception(
                    f"Error in getting columns name from the DeltaTable {self.uri}"
                )
                raise ExternalMetadataSourceException(exception)

        def get_table_names_list(self, database_name: str) -> Iterator[str]:
            """
            Not relevant, just return the current Delta Table URI.

            :param database_name: the database name
            :return: the list of the table names of the database
            """
            yield self.uri

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name of the source.
            """
            return "Delta Table"



from typing import Any, Dict, Iterator, List, Optional, Union

from loguru import logger
from pydantic import Field

from ..metadata_source import ColumnMetadata
from .external_metadata_source import (
    ExternalMetadataSource,
    ExternalMetadataSourceException,
)

try:
    from google.cloud import bigquery

    GCP_INSTALLED = True
except ImportError:
    logger.debug("GCP optional dependency is not installed.")
    GCP_INSTALLED = False

if GCP_INSTALLED:

    class BigQuerySource(ExternalMetadataSource):
        """Instance of a BigQuery source."""

        service_account_json_path: str
        project: Optional[str] = None
        location: Optional[str] = None
        extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

        def create_connection(self) -> None:
            """
            Get the Big Query connection.

            :return:
            """
            try:
                self._connection = bigquery.Client.from_service_account_json(
                    self.service_account_json_path,
                    project=self.project,
                    location=self.location,
                    **self.extra_connection_args,
                )
            except Exception as exception:
                logger.exception("Error when connecting to BigQuery")
                raise exception

        def close_connection(self) -> None:
            """
            Close the BigQuery connection.

            :return:
            """
            self._connection.close()

        def get_column_names(
            self, database_name: str, table_name: str, include_comment: bool = False
        ) -> Iterator[ColumnMetadata]:
            """
            Get column names from the table of the dataset.

            :param database_name: in that case the dataset
            :param table_name: the table name
            :param include_comment: include the comment
            :return: the list of the column names
            """

            try:
                if not self._connection:
                    self.create_connection()

                table_reference = self._connection.dataset(
                    database_name, project=self.project
                ).table(table_name)
                table = self._connection.get_table(table_reference)
                for column in table.schema:
                    column_name = column.name
                    column_comment = None
                    if include_comment and column.description:
                        column_comment = column.description
                    yield ColumnMetadata(
                        column_name=column_name, column_comment=column_comment
                    )
            except Exception as exception:
                logger.exception(
                    f"Error in getting columns name from BigQuery {database_name}.{table_name}"
                )
                raise ExternalMetadataSourceException(exception)

        def get_table_names_list(self, database_name: str) -> Iterator[str]:
            """
            Get the table names list from the GCP dataset.

            :param database_name: in that case the dataset
            :return: the list of the table names list
            """

            try:
                if not self._connection:
                    self.create_connection()
                query_job = self._connection.query(
                    f"SELECT table_name FROM `{database_name}.INFORMATION_SCHEMA.TABLES`"
                )
                results = query_job.result()
                for row in results:
                    yield row.table_name
            except Exception as exception:
                logger.exception(
                    f"Error in getting the table names list name from BigQuery {database_name}"
                )
                raise exception

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name bof the source.
            """
            return "GCP BigQuery"



from .external.aws_source import *
from .external.deltatable_source import *
from .external.external_metadata_source import *
from .external.gcp_source import *
from .external.kafka_schema_registry_source import *
from .external.mysql_source import *
from .external.snowflake_source import *
from .local.avro_schema_source import *
from .local.avro_source import *
from .local.local_metadata_source import *
from .local.orc_source import *
from .local.parquet_source import *
from .metadata_source import *



from unittest.mock import patch

from confluent_kafka.schema_registry import RegisteredSchema, Schema

from metadata_guardian.source import (
    ColumnMetadata,
    KafkaSchemaRegistryAuthentication,
    KafkaSchemaRegistrySource,
)


@patch("confluent_kafka.schema_registry.SchemaRegistryClient")
def test_kafka_schema_registry_source_get_column_names(mock_connection):
    url = "url"
    subject_name = "subject_name"
    expected = [
        ColumnMetadata(column_name="key"),
        ColumnMetadata(column_name="value", column_comment="doc"),
    ]

    source = KafkaSchemaRegistrySource(
        url=url,
    )
    schema_id = "schema_id"
    schema_str = """{
        "fields": [
            {
                "name": "key",
                "type": "string"
            },
            {
                "name": "value",
                "type": "string",
                "doc": "doc"
            }
        ],
        "name": "test_one",
        "namespace": "test.one",
        "type": "record"
    }"""
    schema = RegisteredSchema(
        schema_id=schema_id,
        schema=Schema(schema_str, "AVRO", []),
        subject=subject_name,
        version=1,
    )
    mock_connection.get_latest_version.return_value = schema
    source._connection = mock_connection

    column_names = source.get_column_names(
        database_name=None, table_name=subject_name, include_comment=True
    )

    assert list(column_names) == expected
    assert source.authenticator == KafkaSchemaRegistryAuthentication.USER_PWD


@patch("confluent_kafka.schema_registry.SchemaRegistryClient")
def test_kafka_schema_registry_source_get_table_names_list(mock_connection):
    url = "url"
    expected = ["subject1", "subject2"]

    source = KafkaSchemaRegistrySource(
        url=url,
    )
    subjects = ["subject1", "subject2"]
    mock_connection.get_subjects.return_value = subjects
    source._connection = mock_connection

    subjects_list = source.get_table_names_list(database_name=None)

    assert list(subjects_list) == expected
    assert source.authenticator == KafkaSchemaRegistryAuthentication.USER_PWD



from enum import Enum
from typing import Any, Dict, Iterator, List, Optional

from loguru import logger
from pydantic import Field

from ..metadata_source import ColumnMetadata
from .external_metadata_source import (
    ExternalMetadataSource,
    ExternalMetadataSourceException,
)

try:
    import snowflake.connector
    from snowflake.connector import SnowflakeConnection
    from snowflake.connector.converter_null import SnowflakeNoConverterToPython

    SNOWFLAKE_INSTALLED = True
except ImportError:
    logger.debug("Snowflake optional dependency is not installed.")
    SNOWFLAKE_INSTALLED = False

if SNOWFLAKE_INSTALLED:

    class SnowflakeAuthenticator(Enum):
        """Authentication method for Snowflake source."""

        USER_PWD = 1
        OKTA = 2
        TOKEN = 3

    class SnowflakeSource(ExternalMetadataSource):
        """Instance of a Snowflake source."""

        sf_account: str
        sf_user: str
        sf_password: str
        warehouse: str
        schema_name: str
        okta_account_name: Optional[str] = None
        oauth_token: Optional[str] = None
        oauth_host: Optional[str] = None
        authenticator: SnowflakeAuthenticator = SnowflakeAuthenticator.USER_PWD
        extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

        def create_connection(self) -> None:
            """
            Create a Snowflake connection based on the SnowflakeAuthenticator.

            :return:
            """
            if self.authenticator == SnowflakeAuthenticator.USER_PWD:
                self._connection = snowflake.connector.connect(
                    account=self.sf_account,
                    user=self.sf_user,
                    password=self.sf_password,
                    warehouse=self.warehouse,
                    converter_class=SnowflakeNoConverterToPython,
                    **self.extra_connection_args,
                )
            elif self.authenticator == SnowflakeAuthenticator.OKTA:
                self._connection = snowflake.connector.connect(
                    account=self.sf_account,
                    user=self.sf_user,
                    password=self.sf_password,
                    warehouse=self.warehouse,
                    authenticator=f"https://{self.okta_account_name}.okta.com/",
                    converter_class=SnowflakeNoConverterToPython,
                    **self.extra_connection_args,
                )
            elif self.authenticator == SnowflakeAuthenticator.TOKEN:
                self._connection = snowflake.connector.connect(
                    account=self.sf_account,
                    user=self.sf_user,
                    password=self.sf_password,
                    warehouse=self.warehouse,
                    authenticator="oauth",
                    token=self.oauth_token,
                    converter_class=SnowflakeNoConverterToPython,
                    **self.extra_connection_args,
                )

        def close_connection(self) -> None:
            """
            Close the Snowflake connection.
            :return:
            """
            self._connection.close()

        def get_column_names(
            self, database_name: str, table_name: str, include_comment: bool = False
        ) -> Iterator[ColumnMetadata]:
            """
            Get column names from the table.

            :param database_name: the database name
            :param table_name: the table name
            :param include_comment: include the comment
            :return: the list of the column names
            """
            try:
                if not self._connection or self._connection.is_closed():
                    self.create_connection()
                cursor = self._connection.cursor()
                cursor.execute(
                    f'SHOW COLUMNS IN "{database_name}"."{self.schema_name}"."{table_name}"'
                )
                rows = cursor.fetchall()
                for row in rows:
                    column_name = row[2]
                    column_comment = None
                    if include_comment:
                        column_comment = row[8]
                    yield ColumnMetadata(
                        column_name=column_name, column_comment=column_comment
                    )
            except Exception as exception:
                logger.exception(
                    f"Error in getting columns name from Snowflake {database_name}.{self.schema_name}.{table_name}"
                )
                raise exception
            finally:
                cursor.close()

        def get_table_names_list(self, database_name: str) -> Iterator[str]:
            """
            Get the table names list from the Snowflake database.

            :param database_name: the database name
            :return: the list of the table names of the database
            """
            try:
                if not self._connection or self._connection.is_closed():
                    self.create_connection()
                cursor = self._connection.cursor()
                cursor.execute(f'SHOW TABLES IN DATABASE "{database_name}"')
                rows = cursor.fetchall()
                for row in rows:
                    table_name = row[1]
                    yield table_name.upper()
            except Exception as exception:
                logger.exception(
                    f"Error in getting table names from the database {database_name} in Snowflake"
                )
                raise ExternalMetadataSourceException(exception)
            finally:
                cursor.close()

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name of the source.
            """
            return "Snowflake"



import pytest

from metadata_guardian.source import AvroSchemaSource, ColumnMetadata


@pytest.mark.parametrize(
    "local_file", ["users_avro_schema.json"], indirect=["local_file"]
)
def test_avro_schema_source(local_file):
    source = AvroSchemaSource(local_path=local_file)
    expected = [
        ColumnMetadata(column_name="name"),
        ColumnMetadata(column_name="favorite_number"),
        ColumnMetadata(column_name="favorite_color"),
    ]

    column_names = source.get_column_names()

    assert list(column_names) == expected


@pytest.mark.parametrize(
    "local_file", ["users_avro_schema.json"], indirect=["local_file"]
)
def test_avro_schema_source_namespace(local_file):
    source = AvroSchemaSource(local_path=local_file)
    expected = "example.avro"

    assert expected == source.namespace



import json
from enum import Enum
from typing import Any, Dict, Iterator, List, Optional

from loguru import logger
from pydantic import Field

from ..metadata_source import ColumnMetadata
from .external_metadata_source import (
    ExternalMetadataSource,
    ExternalMetadataSourceException,
)

try:
    from confluent_kafka.schema_registry import SchemaRegistryClient

    KAFKA_SCHEMA_REGISTRY_INSTALLED = True
except ImportError:
    logger.debug("Kafka Schema Registry optional dependency is not installed.")
    KAFKA_SCHEMA_REGISTRY_INSTALLED = False

if KAFKA_SCHEMA_REGISTRY_INSTALLED:

    class KafkaSchemaRegistryAuthentication(Enum):
        """Authentication method for Kafka Schema Registry source."""

        USER_PWD = 1

    class KafkaSchemaRegistrySource(ExternalMetadataSource):
        """Instance of a Kafka Schema Registry source."""

        url: str
        ssl_certificate_location: Optional[str] = None
        ssl_key_location: Optional[str] = None
        authenticator: KafkaSchemaRegistryAuthentication = (
            KafkaSchemaRegistryAuthentication.USER_PWD
        )
        comment_field_name: str = "doc"
        extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

        def create_connection(self) -> None:
            """
            Create the connection of the Kafka Schema Registry.

            :return:
            """
            if self.authenticator == KafkaSchemaRegistryAuthentication.USER_PWD:
                self._connection = SchemaRegistryClient(
                    {"url": self.url, **self.extra_connection_args}
                )
            else:
                raise NotImplementedError()

        def close_connection(self) -> None:
            """
            Close the Kafka Schema Registry connection.

            :return:
            """
            self._connection.__exit__()

        def get_column_names(
            self, database_name: str, table_name: str, include_comment: bool = False
        ) -> Iterator[ColumnMetadata]:
            """
            Get the column names from the subject.

            :param database_name: not relevant
            :param table_name: the subject name
            :param include_comment: include the comment
            :return: the list of the column names
            """
            try:
                if not self._connection:
                    self.create_connection()
                registered_schema = self._connection.get_latest_version(table_name)
                for field in json.loads(registered_schema.schema.schema_str)["fields"]:
                    column_name = field["name"]
                    column_comment = None
                    if (
                        include_comment
                        and self.comment_field_name in field
                        and field[self.comment_field_name]
                    ):
                        column_comment = field[self.comment_field_name]
                    yield ColumnMetadata(
                        column_name=column_name, column_comment=column_comment
                    )
            except Exception as exception:
                logger.exception(
                    f"Error in getting columns name from the Kafka Schema Registry {table_name}"
                )
                raise exception

        def get_table_names_list(self, database_name: str) -> Iterator[str]:
            """
            Get all the subjects from the Schema Registry.

            :param database_name: not relevant in that case
            :return: the list of the table names of the database
            """
            try:
                if not self._connection:
                    self.create_connection()
                yield from self._connection.get_subjects()
            except Exception as exception:
                logger.exception(
                    f"Error all the subjects from the subject in the Kafka Schema Registry"
                )
                raise ExternalMetadataSourceException(exception)

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name of the source.
            """
            return "Kafka Schema Registry"



import argparse
import os

from metadata_guardian import AvailableCategory, ColumnScanner, DataRules
from metadata_guardian.source import AvroSource, ORCSource, ParquetSource


def get_gcp_bigquery() -> ExternalMetadataSource:
    return BigQuerySource(
        service_account_json_path=os.environ["BIGQUERY_SERVICE_ACCOUNT"],
        project=os.environ["BIGQUERY_PROJECT"],
        location=os.environ["BIGQUERY_LOCATION"],
    )


def get_kafka_schema_registry() -> ExternalMetadataSource:
    return KafkaSchemaRegistrySource(url=os.environ["KAFKA_SCHEMA_REGISTRY_URL"])


def get_delta_table() -> ExternalMetadataSource:
    return DeltaTableSource(uri=os.environ["DELTA_TABLE_URI"])


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-rules",
        choices=["PII", "INCLUSION"],
        default="PII",
        help="The Data Rules to use",
    )
    parser.add_argument(
        "--local-source",
        choices=["Avro", "Parquet", "Orc"],
        required=True,
        help="The Local Metadata Source to use",
    )
    parser.add_argument(
        "--scanner", choices=["ColumnScanner"], help="The scanner to use"
    )
    parser.add_argument("--path", required=True, help="The path of the file to scan")
    parser.add_argument(
        "--include_comments", default=True, help="Include the comments in the scan"
    )
    args = parser.parse_args()
    data_rules = DataRules.from_available_category(
        category=AvailableCategory[args.data_rules]
    )
    column_scanner = ColumnScanner(
        data_rules=data_rules, progression_bar_disabled=False
    )

    if args.local_source == "Avro":
        source = AvroSource(local_path=args.path)
    elif args.local_source == "Parquet":
        source = ParquetSource(local_path=args.path)
    elif args.local_source == "Orc":
        source = ORCSource(local_path=args.path)

    report = column_scanner.scan_local(source)
    report.to_console()



from enum import Enum
from typing import Any, Dict, Iterator, List, Optional

from loguru import logger
from pydantic import Field

from ..metadata_source import ColumnMetadata
from .external_metadata_source import (
    ExternalMetadataSource,
    ExternalMetadataSourceException,
)

try:
    import pymysql

    MYSQL_INSTALLED = True
except ImportError:
    logger.debug("MySQL optional dependency is not installed.")
    MYSQL_INSTALLED = False

if MYSQL_INSTALLED:

    class MySQLAuthenticator(Enum):
        """Authentication method for MySQL source."""

        USER_PWD = 1

    class MySQLSource(ExternalMetadataSource):
        """Instance of a MySQL source."""

        user: str
        password: str
        host: str
        database: Optional[str] = None
        authenticator: MySQLAuthenticator = MySQLAuthenticator.USER_PWD
        extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

        def create_connection(self) -> None:
            """
            Create a MySQL connection based on the MySQLAuthenticator.

            :return:
            """
            if self.authenticator == MySQLAuthenticator.USER_PWD:
                self._connection = pymysql.connect(
                    host=self.host,
                    user=self.user,
                    password=self.password,
                    database=self.database,
                    cursorclass=pymysql.cursors.DictCursor,
                    **self.extra_connection_args,
                )

        def get_column_names(
            self, database_name: str, table_name: str, include_comment: bool = False
        ) -> Iterator[ColumnMetadata]:
            """
            Get column names from the table.

            :param database_name: the database name
            :param table_name: the table name
            :param include_comment: include the comment
            :return: the list of the column names
            """
            try:
                if not self._connection or not self._connection.open:
                    self.create_connection()
                cursor = self._connection.cursor()
                cursor.execute(f"SHOW FULL COLUMNS FROM {database_name}.{table_name}")
                rows = cursor.fetchall()
                for row in rows:
                    column_name = row["Field"]
                    column_comment = None
                    if include_comment and row["Comment"]:
                        column_comment = row["Comment"]
                    yield ColumnMetadata(
                        column_name=column_name, column_comment=column_comment
                    )
            except Exception as exception:
                logger.exception(
                    f"Error in getting columns name from MySQL {database_name}.{table_name}"
                )
                raise exception
            finally:
                cursor.close()

        def get_table_names_list(self, database_name: str) -> Iterator[str]:
            """
            Get the table names list from the MySQL database.

            :param database_name: the database name
            :return: the list of the table names of the database
            """
            try:
                if not self._connection or not self._connection.open:
                    self.create_connection()
                cursor = self._connection.cursor()
                cursor.execute(f"SHOW TABLES IN {database_name}")
                rows = cursor.fetchall()
                for row in rows:
                    table_name = list(row.values())[0]
                    yield table_name
            except Exception as exception:
                logger.exception(
                    f"Error in getting table names from the database {database_name} in MySQL"
                )
                raise ExternalMetadataSourceException(exception)
            finally:
                cursor.close()

        @classmethod
        def type(cls) -> str:
            """
            The type of the source.

            :return: the name of the source.
            """
            return "MySQL"



import pytest

from metadata_guardian.source import ColumnMetadata
from metadata_guardian.source.local.avro_source import AvroSource


@pytest.mark.parametrize("local_file", ["users.avro"], indirect=["local_file"])
def test_avro_source(local_file):
    source = AvroSource(local_path=local_file)
    expected = [
        ColumnMetadata(column_name="name"),
        ColumnMetadata(column_name="favorite_number"),
        ColumnMetadata(column_name="favorite_color"),
    ]

    column_names = source.get_column_names()

    assert list(column_names) == expected


@pytest.mark.parametrize("local_file", ["users.avro"], indirect=["local_file"])
def test_avro_source_namespace(local_file):
    source = AvroSource(local_path=local_file)
    expected = "example.avro"

    assert expected == source.namespace


@pytest.mark.parametrize("local_file", ["users.avro"], indirect=["local_file"])
def test_avro_source_attribute_type(local_file):
    source = AvroSource(local_path=local_file)
    expected = [
        ColumnMetadata(column_name="string"),
        ColumnMetadata(column_name="['int', 'null']"),
        ColumnMetadata(column_name="['string', 'null']"),
    ]

    field_attribute_type = source.get_field_attribute(attribute_name="type")

    assert expected == field_attribute_type



from typing import Any, Dict, Iterator, List

import pyarrow
from pyarrow.dataset import Dataset
from pyarrow.fs import FileSystem, LocalFileSystem
from pydantic import Field

from ..metadata_source import ColumnMetadata, MetadataSource


class LocalMetadataSource(MetadataSource):
    """LocalMetadata Source contract."""

    local_path: str
    fs: FileSystem = LocalFileSystem()
    extra_connection_args: Dict[str, Any] = Field(default_factory=dict)

    class Config:
        arbitrary_types_allowed = True

    def read(self) -> Dataset:
        """
        Read the source local file.

        :return: the file content
        """
        return pyarrow.dataset.dataset(
            self.local_path, filesystem=self.fs, **self.extra_connection_args
        )

    def get_column_names(self) -> Iterator[ColumnMetadata]:
        """
        Get the column names from the schema.

        :return: the list of the column names
        """
        for column_name in self.read().schema.names:
            yield ColumnMetadata(column_name=column_name)
